@@ -2,6 +2,7 @@ from collections import deque, defaultdict
 from contextlib import contextmanager
 import copy
 import csv
+import json
 import math
 import os
 import pickle
@@ -116,11 +117,45 @@ class KB():
         effects)
 
         :Example:
+        
         >>> kb = KB()
         >>> kb.set_propagation_limit(1)
         """
         self._PROPAGATION_LIMIT = propagations
 
+    def nodes(self, filter_fn=None):
+        """Returns the nodes in the current KB, optionally filtered by filter_fn.
+        
+        :param filter_fn: Function which is passed each of the nodes; only
+        nodes for which it returns True will be returned.
+        
+        :Example:
+
+        >>> kb = KB()
+        >>> kb.store('node_friends(chris, jonny)')
+        0
+        >>> list(kb.nodes())
+        [chris, jonny]
+        >>> chris_node = list(kb.nodes(lambda x: x == 'chris'))
+        >>> len(chris_node)
+        1
+        >>> chris_node[0] == 'chris'
+        True
+        >>> kb.node('jonny').passenger_age = 14
+        >>> jonny_node = list(kb.nodes(lambda x: x.passenger_age == 14))
+        >>> jonny_node[0] == 'jonny'
+        True
+        
+        """
+        nodes = self.G.nodes(data=True)
+        for node_name, node_attrs in nodes:
+            node = self.node(node_name)
+            if filter_fn:
+                if filter_fn(node):
+                    yield node
+            else:
+                yield node
+
     def edge(self, sub, pred, ob):
         """Returns an edge and its attributes.
 
@@ -146,6 +181,31 @@ class KB():
             self._edge_cache[(sub, pred, ob)] = edge
         return edge
     
+    def edges(self, filter_fn=None):
+        """Returns edges in the KB, optionally filtered by filter_fn.
+
+        :Example:
+
+        >>> kb = KB()
+        >>> kb.store('eats(tom, rice)')
+        0
+        >>> list(kb.edges())
+        [tom___eats___rice]
+        >>> list(kb.edges(lambda x: x.nodes[0] == 'tom'))
+        [tom___eats___rice]
+        >>> kb.edge('tom', 'eats', 'rice').alot = 'every_day_almost'
+        >>> list(kb.edges(lambda x: x.alot == 'every_day_almost'))
+        [tom___eats___rice]
+        """
+        edges = self.G.edges(data=True)
+        for edge in edges:
+            edge = self.edge(edge[0], edge[-1]['pred'], edge[1])
+            if filter_fn:
+                if filter_fn(edge):
+                    yield edge
+            else:
+                yield edge
+    
     @contextmanager
     def dont_propagate(self):
         self._dont_propagate = True
@@ -969,7 +1029,46 @@ class KB():
         for (u, p, v) in triples:
             self.store('{}({},{})'.format(p, u, v))
 
+    def to_csv(self, csvfile, delimiter=','):
+        """Saves a knowledge base to a CSV file.
+
+        :param str csvfile: Filename to write. File will be a number of rows separated
+        by delimiter, in the 6-column format `subject,predicate,object,subject_attributes
+        ,object_attributes,edge_attributes`.
+        :param str delimiter: Default is `','`; try `'\t'` for tab delimited.
+        """
+        with open(csvfile, 'w') as f:
+            f = csv.writer(f)
+            for rule in self.rules:
+                rule = split_to_parts(str(rule.head))
+                _sub = self.node(rule[0])
+                _ob = self.node(rule[2])
+                _sub_attrs = json.dumps(_sub.attrs)
+                _ob_attrs = json.dumps(_ob.attrs)
+                _edge_attrs = json.dumps(self.edge(str(_sub), rule[1], str(_ob)).attrs)
+                f.writerow([_sub, rule[1], _ob, _sub_attrs, _ob_attrs, _edge_attrs])
+
+
     def from_csv(self, csvfile, header=None, start=0, size=None, delimiter=','):
+        """Reads a knowledge base into memory from a CSV file.
+
+        :param str csvfile: Filename to read. File is a number of rows separated
+        by delimiter, in the basic format `subject,predicate,object`. Optionally,
+        the row may have 3 more columns: `subject_attributes`, `object_attributes`
+        and `edge_attributes`.
+        :param bool header: Whether to ignore (True) a header row
+        :param int start: Aside from the header row, number of subsequent rows to ignore
+        :param int size: How many rows after which to stop reading, or None if never.
+        :param str delimiter: Default is `','`; try `'\t'` for tab delimited.
+
+        :Example:
+        
+        >>> kb = KB()
+        >>> kb.from_csv('./assets/countries_s1_train.csv', delimiter='\\t')
+        >>> list(kb.query('locatedin(fiji, Where)'))
+        [{'Where': 'melanesia'}, {'Where': 'oceania'}]
+
+        """
         with open(csvfile) as f:
             reader = csv.reader(f, delimiter=delimiter)
             i = 0
@@ -986,13 +1085,12 @@ class KB():
                 ob = cleanse(row[2])
                 if not (sub.replace('_','').isalnum() and ob.replace('_','').isalnum()):
                     continue
-                node_attributes = [
-                    {'_': {'real_name': row[0]}},
-                    {'_': {'real_name': row[2]}}
-                ]
-                edge_attributes = {
-                    '_': {'real_name': row[1]}
-                }
+                if len(row) > 3:
+                    node_attributes = [json.loads(row[3]), json.loads(row[4])]
+                    edge_attributes = json.loads(row[5])
+                else:
+                    node_attributes = []
+                    edge_attributes = {}
                 self.store('{}({},{})'.format(pred, sub, ob), node_attributes=node_attributes, edge_attributes=edge_attributes)
                 i += 1
                 if size and i > size:
